import torchvision

from .builder import SURROGATE_REGISTRY

weights_dict = dict(
    vgg16=torchvision.models.VGG16_Weights.IMAGENET1K_V1,
    vgg19=torchvision.models.VGG19_Weights.IMAGENET1K_V1,
    resnet50=torchvision.models.ResNet50_Weights.IMAGENET1K_V1,
    resnet152=torchvision.models.ResNet152_Weights.IMAGENET1K_V1,
    densenet121=torchvision.models.DenseNet121_Weights.IMAGENET1K_V1,
    densenet169=torchvision.models.DenseNet169_Weights.IMAGENET1K_V1,
    inception_v3=torchvision.models.Inception_V3_Weights.IMAGENET1K_V1,
    vit_b_16=torchvision.models.ViT_B_16_Weights.IMAGENET1K_V1,
    vit_b_32=torchvision.models.ViT_B_32_Weights.IMAGENET1K_V1,
    swin_b=torchvision.models.Swin_B_Weights.IMAGENET1K_V1)


@SURROGATE_REGISTRY.register()
def inception_v3(pretrain=True):
    return torchvision.models.inception_v3(
        weights=weights_dict['inception_v3'] if pretrain else None)


@SURROGATE_REGISTRY.register()
def vgg16(pretrain=True):
    return torchvision.models.vgg16(
        weights=weights_dict['vgg16'] if pretrain else None)


@SURROGATE_REGISTRY.register()
def vgg19(pretrain=True):
    return torchvision.models.vgg19(
        weights=weights_dict['vgg19'] if pretrain else None)


@SURROGATE_REGISTRY.register()
def resnet50(pretrain=True):
    return torchvision.models.resnet50(
        weights=weights_dict['resnet50'] if pretrain else None)


@SURROGATE_REGISTRY.register()
def resnet152(pretrain=True):
    return torchvision.models.resnet152(
        weights=weights_dict['resnet152'] if pretrain else None)


@SURROGATE_REGISTRY.register()
def densenet121(pretrain=True):
    return torchvision.models.densenet121(
        weights=weights_dict['densenet121'] if pretrain else None)


@SURROGATE_REGISTRY.register()
def densenet169(pretrain=True):
    return torchvision.models.densenet169(
        weights=weights_dict['densenet169'] if pretrain else None)


@SURROGATE_REGISTRY.register()
def vit_b_16(pretrain=True):
    return torchvision.models.vit_b_32(
        weights=weights_dict['vit_b_32'] if pretrain else None)


@SURROGATE_REGISTRY.register()
def vit_b_32(pretrain=True):
    return torchvision.models.vit_b_32(
        weights=weights_dict['vit_b_32'] if pretrain else None)


@SURROGATE_REGISTRY.register()
def swin_b(pretrain=True):
    return torchvision.models.swin_b(
        weights=weights_dict['swin_b'] if pretrain else None)



import torch
from torch import nn


class EnhancedBN(nn.Module):
    def __init__(self, nc: int, sty_nc: int = 3, sty_nhidden: int = 128):
        super(EnhancedBN, self).__init__()
        self.bn = nn.BatchNorm2d(nc)
        self.mapping = nn.Conv2d(
            in_channels=sty_nc,
            out_channels=sty_nhidden,
            kernel_size=3,
            padding=1,
            stride=1,
        )
        self.gamma = nn.Conv2d(
            in_channels=sty_nhidden,
            out_channels=nc,
            kernel_size=3,
            padding=1,
            stride=1,
        )
        self.beta = nn.Conv2d(
            in_channels=sty_nhidden,
            out_channels=nc,
            kernel_size=3,
            padding=1,
            stride=1,
        )
        self.init_weight()

    def init_weight(self):
        nn.init.kaiming_normal_(self.mapping.weight)
        nn.init.kaiming_normal_(self.gamma.weight)
        nn.init.kaiming_normal_(self.beta.weight)

    def forward(self, base, sty):
        bn = self.bn(base)
        sty_resized = torch.nn.functional.interpolate(
            sty, size=bn.size()[2:], mode='bilinear'
        )
        actv = torch.nn.functional.relu(self.mapping(sty_resized))
        # style injection
        bn = bn * (1 + self.gamma(actv)) + self.beta(actv)
        return bn


class ResidualBlock(nn.Module):
    def __init__(self, num_filters):
        super(ResidualBlock, self).__init__()
        self.block1 = nn.Sequential(
            nn.ReflectionPad2d(1),
            nn.Conv2d(
                in_channels=num_filters,
                out_channels=num_filters,
                kernel_size=3,
                stride=1,
                padding=0,
                bias=False,
            ),
        )
        self.bn1 = EnhancedBN(num_filters)
        self.block2 = nn.Sequential(
            nn.ReLU(True),
            nn.Dropout(0.5),
            nn.ReflectionPad2d(1),
            nn.Conv2d(
                in_channels=num_filters,
                out_channels=num_filters,
                kernel_size=3,
                stride=1,
                padding=0,
                bias=False,
            ),
        )
        self.bn2 = EnhancedBN(num_filters)

    def forward(self, x, sty):
        residual = self.block1(x)
        residual = self.bn1(residual, sty)
        residual = self.block2(residual)
        residual = self.bn2(residual, sty)
        return x + residual


ngf = 64


class ResNetGenerator(nn.Module):
    def __init__(self):
        super(ResNetGenerator, self).__init__()
        self.block1 = nn.Sequential(
            nn.ReflectionPad2d(3),
            nn.Conv2d(3, ngf, kernel_size=7, padding=0, bias=False),
        )
        self.bn1 = EnhancedBN(ngf)
        # Input size = 3, n, n
        self.block2 = nn.Sequential(
            nn.Conv2d(
                ngf, ngf * 2, kernel_size=3, stride=2, padding=1, bias=False
            ),
        )
        self.bn2 = EnhancedBN(ngf * 2)
        # Input size = 3, n/2, n/2
        self.block3 = nn.Sequential(
            nn.Conv2d(
                ngf * 2,
                ngf * 4,
                kernel_size=3,
                stride=2,
                padding=1,
                bias=False,
            ),
        )
        self.bn3 = EnhancedBN(ngf * 4)
        # Input size = 3, n/4, n/4
        # Residual Blocks: 6
        self.resblock1 = ResidualBlock(ngf * 4)
        self.resblock2 = ResidualBlock(ngf * 4)
        self.resblock3 = ResidualBlock(ngf * 4)
        self.resblock4 = ResidualBlock(ngf * 4)
        self.resblock5 = ResidualBlock(ngf * 4)
        self.resblock6 = ResidualBlock(ngf * 4)
        # Input size = 3, n/4, n/4
        self.upsampl1 = nn.ConvTranspose2d(
            ngf * 4,
            ngf * 2,
            kernel_size=3,
            stride=2,
            padding=1,
            output_padding=1,
            bias=False,
        )
        self.ubn1 = EnhancedBN(ngf * 2)
        # Input size = 3, n/2, n/2
        self.upsampl2 = nn.ConvTranspose2d(
            ngf * 2,
            ngf,
            kernel_size=3,
            stride=2,
            padding=1,
            output_padding=1,
            bias=False,
        )
        self.ubn2 = EnhancedBN(ngf)
        # Input size = 3, n, n
        self.blockf = nn.Sequential(
            nn.ReflectionPad2d(3), nn.Conv2d(ngf, 3, kernel_size=7, padding=0)
        )

    def forward(self, input, sty):
        x = self.block1(input)
        x = self.bn1(x, sty)
        x = torch.nn.functional.relu(x)
        x = self.block2(x)
        x = self.bn2(x, sty)
        x = torch.nn.functional.relu(x)
        x = self.block3(x)
        x = self.bn3(x, sty)
        x = torch.nn.functional.relu(x)
        # =============================
        x = self.resblock1(x, sty)
        x = self.resblock2(x, sty)
        x = self.resblock3(x, sty)
        x = self.resblock4(x, sty)
        x = self.resblock5(x, sty)
        x = self.resblock6(x, sty)
        # =============================
        x = self.upsampl1(x)
        x = self.ubn1(x, sty)
        x = torch.nn.functional.relu(x)
        x = self.upsampl2(x)
        x = self.ubn2(x, sty)
        x = torch.nn.functional.relu(x)
        x = self.blockf(x)
        return (torch.tanh(x) + 1) / 2


AIMGenerator = ResNetGenerator



from ...runtime.factory import Registry

SURROGATE_REGISTRY = Registry('SURROGATE')


def build_surrogate(_type: str, *args, **kwargs) -> object:
    return SURROGATE_REGISTRY.get(_type)(*args, **kwargs)


def list_surrogates():
    return list(SURROGATE_REGISTRY._obj_map.keys())



from .builder import build_surrogate, list_surrogates
from .hooks import midlayer_dict, register_collecter
from .tv import (densenet121, densenet169, inception_v3, resnet50, resnet152,
                 swin_b, vgg16, vgg19, vit_b_16, vit_b_32)

__all__ = [
    'build_surrogate', 'list_surrogates', 'inception_v3', 'vgg16', 'vgg19',
    'resnet50', 'resnet152', 'densenet121', 'densenet169', 'vit_b_16',
    'vit_b_32', 'swin_b', 'midlayer_dict', 'register_collecter'
]



import torch

midlayer_dict = {
    'vgg16': 'features.16',
    'vgg19': 'features.18',
    'resnet152': 'layer2',
    'densenet169': 'features.denseblock2',
    'inception_v3': 'Mixed_6c',
}

feat_collecter = []


def register_collecter(m: torch.nn.Module, layer: str):

    def _hook(m, i, o):
        feat_collecter.append(o)

    _handler = m.get_submodule(layer).register_forward_hook(_hook)
    return _handler, feat_collecter



from collections import OrderedDict
from pathlib import Path
from typing import Union

import torch

from .generator.aim import AIMGenerator


class AIMAttack:

    def __init__(self,
                 device: Union[str, torch.device],
                 epsilon: float = 16 / 255) -> None:
        if isinstance(device, str):
            device = torch.device(device)
        self.device = device
        self.adv_gen = AIMGenerator().to(device)
        self.set_mode('eval')
        self.epsilon = epsilon

    def load_ckpt(self, ckpt: Union[str, Path, OrderedDict]) -> None:
        if isinstance(ckpt, str):
            ckpt = Path(ckpt)
        if isinstance(ckpt, Path):
            if not ckpt.exists():
                raise FileNotFoundError(f'File not found: {ckpt}')
            ckpt = torch.load(ckpt, map_location=self.device)
        self.adv_gen.load_state_dict(ckpt)
        self.adv_gen.to(self.device)

    def save_ckpt(self, ckpt: Union[str, Path]) -> None:
        if isinstance(ckpt, str):
            ckpt = Path(ckpt)
        _adv_gen_cpu = self.adv_gen.to('cpu')
        torch.save(_adv_gen_cpu.state_dict(), ckpt)

    def get_params(self) -> torch.nn.Parameter:
        return self.adv_gen.parameters()

    def set_mode(self, mode: str) -> None:
        assert mode in ['train', 'eval']
        self.adv_gen.train() if mode == 'train' else self.adv_gen.eval()

    def __call__(self, x_nat, x_guid) -> torch.Tensor:
        x_adv = self.adv_gen(x_nat, x_guid)
        x_adv = torch.min(torch.max(x_adv, x_nat - self.epsilon),
                          x_nat + self.epsilon)
        torch.clamp_(x_adv, 0.0, 1.0)
        return x_adv



from typing import List

import torchvision

from . import env


def resize_256_224() -> List:
    return [
        torchvision.transforms.Resize(size=256),
        torchvision.transforms.CenterCrop(size=(224, 224)),
    ]


def resize_512_448() -> List:
    return [
        torchvision.transforms.Resize(size=512),
        torchvision.transforms.CenterCrop(size=(448, 448)),
    ]


def resize_224() -> List:
    return [torchvision.transforms.Resize(size=224)]


def hflip() -> List:
    return [torchvision.transforms.RandomHorizontalFlip(0.5)]


def to_ts() -> List:
    return [torchvision.transforms.ToTensor()]


def to_pil() -> List:
    return [torchvision.transforms.ToPILImage()]


def to_color() -> List:
    return [
        torchvision.transforms.Lambda(lambda x: x.convert('RGB')
                                      if x.mode != 'RGB' else x)
    ]


def norm(dataset: str = 'IMAGENET', _callable: bool = False) -> List:
    dataset = dataset.upper()
    assert dataset in ['CIFAR10', 'CIFAR100', 'IMAGENET']
    mean_std = (
        getattr(env, dataset + '_DEFAULT_MEAN'),
        getattr(env, dataset + '_DEFAULT_STD'),
    )
    transforms = [torchvision.transforms.Normalize(*mean_std)]
    if _callable:
        return torchvision.transforms.Compose(transforms)
    return transforms



#!/usr/bin/env python3
import sys
from pathlib import Path
from typing import Union

from setuptools import find_packages, setup

try:
    sys.path.append(str(Path(__file__).parent / 'src'))
    import gat
except ImportError:
    print('Please install aim_attack package')
    sys.exit(1)


def read(path: Union[str, Path]) -> str:
    with open(path, 'r') as f:
        return f.read()


setup(name='pygat',
      version=gat.version(),
      description='GAT: Generative Attack Toolbox',
      long_description=read('README.md'),
      long_description_content_type='text/markdown',
      author='Terry Li',
      url='https://terrytengli.com/gat/',
      classifiers=[
          'Programming Language :: Python :: 3',
          'License :: OSI Approved :: MIT License'
      ],
      python_requires='>=3.10',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      install_requires=['tqdm', 'tabulate', 'torch', 'torchvision'],
      include_package_data=True)



import torch


class ContrastiveLoss(torch.nn.Module):
    """
    Contrastive loss
    Adapted from: (OnlineContrastiveLoss)
        https://github.com/adambielski/siamese-triplet/blob/master/losses.py
    Based on: http://yann.lecun.com/exdb/publis/pdf/hadsell-chopra-lecun-06.pdf
    """

    def __init__(self, margin):
        super(ContrastiveLoss, self).__init__()
        self.margin = margin

    def forward(self, anchors, negatives, positives):
        anchors = anchors / anchors.norm(dim=-1, keepdim=True)
        negatives = negatives / negatives.norm(dim=-1, keepdim=True)
        positives = positives / positives.norm(dim=-1, keepdim=True)

        positive_loss = (anchors - positives).pow(2).sum(1)
        negative_loss = torch.nn.functional.relu(
            self.margin - (anchors - negatives).pow(2).sum(1).sqrt()).pow(2)

        loss = 0.5 * torch.cat([positive_loss, negative_loss], dim=0)

        return loss.mean()



import random

import numpy as np
import torch


def fix_random(seed: int = 0) -> None:
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def randid(k: int = 4):
    charset = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(random.choices(charset, k=k))



# Code is copied from
# github.com/Alibaba-AAIG/Beyond-ImageNet-Attack:generator.py@863b7
import torch
from torch import nn


class ResidualBlock(nn.Module):
    def __init__(self, num_filters):
        super(ResidualBlock, self).__init__()
        self.block = nn.Sequential(
            nn.ReflectionPad2d(1),
            nn.Conv2d(
                in_channels=num_filters,
                out_channels=num_filters,
                kernel_size=3,
                stride=1,
                padding=0,
                bias=False,
            ),
            nn.BatchNorm2d(num_filters),
            nn.ReLU(True),
            nn.Dropout(0.5),
            nn.ReflectionPad2d(1),
            nn.Conv2d(
                in_channels=num_filters,
                out_channels=num_filters,
                kernel_size=3,
                stride=1,
                padding=0,
                bias=False,
            ),
            nn.BatchNorm2d(num_filters),
        )

    def forward(self, x):
        residual = self.block(x)
        return x + residual


ngf = 64


class ResNetGenerator(nn.Module):
    """
    https://github.com/Alibaba-AAIG/Beyond-ImageNet-Attack/blob/863b758ee4f4a6d3d4e7777c5f94f457fa449f73/generator.py#L14

    Test Case:
    >>> netG = ResNetGenerator()
    >>> test_sample = torch.rand(1, 3, 224, 224)
    >>> print("Generator output:", netG(test_sample).size())
    >>> print(
    >>>     "Generator parameters:",
    >>>     sum(p.numel() for p in netG.parameters() if p.requires_grad),
    >>> )
    """

    def __init__(self, inception=False):
        super(ResNetGenerator, self).__init__()
        self.inception = inception
        self.block1 = nn.Sequential(
            nn.ReflectionPad2d(3),
            nn.Conv2d(3, ngf, kernel_size=7, padding=0, bias=False),
            nn.BatchNorm2d(ngf),
            nn.ReLU(True),
        )
        # output: (ngf) x (n) x (n)
        self.block2 = nn.Sequential(
            nn.Conv2d(
                ngf, ngf * 2, kernel_size=3, stride=2, padding=1, bias=False
            ),
            nn.BatchNorm2d(ngf * 2),
            nn.ReLU(True),
        )
        # output: (ngf*2) x (n/2) x (n/2)
        self.block3 = nn.Sequential(
            nn.Conv2d(
                ngf * 2,
                ngf * 4,
                kernel_size=3,
                stride=2,
                padding=1,
                bias=False,
            ),
            nn.BatchNorm2d(ngf * 4),
            nn.ReLU(True),
        )
        # output: (ngf*4) x (n/4) x (n/4)
        self.resblock1 = ResidualBlock(ngf * 4)
        self.resblock2 = ResidualBlock(ngf * 4)
        self.resblock3 = ResidualBlock(ngf * 4)
        self.resblock4 = ResidualBlock(ngf * 4)
        self.resblock5 = ResidualBlock(ngf * 4)
        self.resblock6 = ResidualBlock(ngf * 4)
        # output: (ngf*4) x (n/4) x (n/4)
        self.upsampl1 = nn.Sequential(
            nn.ConvTranspose2d(
                ngf * 4,
                ngf * 2,
                kernel_size=3,
                stride=2,
                padding=1,
                output_padding=1,
                bias=False,
            ),
            nn.BatchNorm2d(ngf * 2),
            nn.ReLU(True),
        )
        # output: (ngf*2) x (n/2) x (n/2)
        self.upsampl2 = nn.Sequential(
            nn.ConvTranspose2d(
                ngf * 2,
                ngf,
                kernel_size=3,
                stride=2,
                padding=1,
                output_padding=1,
                bias=False,
            ),
            nn.BatchNorm2d(ngf),
            nn.ReLU(True),
        )
        # output: (ngf) x (n) x (n)
        self.blockf = nn.Sequential(
            nn.ReflectionPad2d(3), nn.Conv2d(ngf, 3, kernel_size=7, padding=0)
        )
        self.crop = nn.ConstantPad2d((0, -1, -1, 0), 0)

    def forward(self, input):
        x = self.block1(input)
        x = self.block2(x)
        x = self.block3(x)
        x = self.resblock1(x)
        x = self.resblock2(x)
        x = self.resblock3(x)
        x = self.resblock4(x)
        x = self.resblock5(x)
        x = self.resblock6(x)
        x = self.upsampl1(x)
        x = self.upsampl2(x)
        x = self.blockf(x)
        if self.inception:
            x = self.crop(x)
        return (torch.tanh(x) + 1) / 2


CDAGenerator = ResNetGenerator



from pathlib import Path
from typing import List, Union

import torch
import torchvision

from .builder import DATASET_REGISTRY
from .transforms import resize_256_224, to_color, to_ts


@DATASET_REGISTRY.register()
def imagenet(
    data_root: Union[str, Path],
    is_train: bool = True,
    filter_class: Union[int, List[int]] = None,
) -> torch.utils.data.Dataset:
    if isinstance(data_root, str):
        data_root = Path(data_root)
    if is_train:
        data_root = data_root / 'train'
    else:
        data_root = data_root / 'val'

    _transforms = resize_256_224() + to_color() + to_ts()

    _ds = torchvision.datasets.ImageFolder(
        data_root,
        transform=torchvision.transforms.Compose(_transforms),
    )

    if isinstance(filter_class, int):
        filter_class = [filter_class]
    if filter_class:
        _ds.samples = list(filter(lambda x: x[1] in filter_class, _ds.samples))

    return _ds
